from __future__ import print_function
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import os
import os.path as osp
import math
import time
import data_loader
import resnet as models
import alexnet_1 as models

import warnings
warnings.filterwarnings('ignore')

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
iteration = 30000
momentum = 0.9
l2_decay = 5e-4
cuda = True

class HLoss(nn.Module):
    def __init__(self):
        super(HLoss, self).__init__()

    def forward(self, x):
        #b = F.softmax(x, dim=1) * F.log_softmax(x, dim=1)
        b = x * torch.log(x)
        b = -1.0 * b.sum()
        return b

def softmax(data):
    ''' Element-wise normalization over a list of FloatTensors
        Arguments:
            - data: list of FloatTensors with the same length
    '''
    data_exp = [ torch.exp(x) for x in data ]
    return [ x / sum(data_exp) for x in data_exp ]

def mahalanobis_metric(p, S, U):
    r''' Compute the mahalanobis distance between the encoding of a sample (p) and a set (S).
    Args:
        p: tensor (batch_size, dim), a batch of samples features
        S: tensor (size, dim), a domain which contains a set of samples features
    Return:
        mahalanobis_distances: tensor (batch_size)
    '''
    mu_S = torch.mean(S, dim=0, keepdim=True) # (1, dim)
    mahalanobis_distances = (p - mu_S).mm(U.mm(U.t())).mm((p - mu_S).t())
    return mahalanobis_distances.diag().sqrt().clamp(0.01, 2)

def domain_encoding(model,source1_loader,source2_loader):
    model.eval()
    with torch.no_grad():
        S1 = []
        S2 = []
        for batch, label in source1_loader:
            if cuda:
                batch = Variable(batch.cuda())
            data1 = model.sharedNet(batch)
            #fea_son1 = model.bottleneck(data1)
            fea_son1 = model.sonnet1(data1)
            fea_son1 = model.avgpool(fea_son1)
            fea_son1 = fea_son1.view(fea_son1.size(0), -1)
            S1.append(fea_son1)
        for batch, label in source2_loader:
            if cuda:
                batch = Variable(batch.cuda())
            data2 = model.sharedNet(batch)
            #fea_son2 = model.bottleneck(data2)
            fea_son2 = model.sonnet2(data2)
            fea_son2 = model.avgpool(fea_son2)
            fea_son2 = fea_son2.view(fea_son2.size(0), -1)
            S2.append(fea_son2)
        S1 = torch.cat(S1, 0)
        S2 = torch.cat(S2, 0)
        mu_S1 = torch.mean(S1, dim=0, keepdim=True)
        mu_S2 = torch.mean(S2, dim=0, keepdim=True)
        return mu_S1,mu_S2

def train(source1,source2,target,model,Us,source1_loader,source2_loader,target_train_loader,target_test_loader,log_interval,lr):
    source1_iter = iter(source1_loader)
    source2_iter = iter(source2_loader)
    target_iter = iter(target_train_loader)
    correct = 0
    moe_criterion = nn.NLLLoss()
    entropy_criterion = HLoss()

    for i in range(1, iteration + 1):
        model.train()
        LEARNING_RATE = lr / math.pow((1 + 10 * (i - 1) / (iteration)), 0.75)# 学习率衰减
        if (i - 1) % 100 == 0:
            print("learning rate：", LEARNING_RATE)
        optimizer = torch.optim.SGD([
            {'params': model.sharedNet.parameters()},
            #{'params': model.bottleneck.parameters(), 'lr': LEARNING_RATE},
            {'params': model.sonnet1.parameters(), 'lr': LEARNING_RATE},
            {'params': model.sonnet2.parameters(), 'lr': LEARNING_RATE},
            {'params': model.classifier_layer_son1_1.parameters(), 'lr': LEARNING_RATE},
            {'params': model.classifier_layer_son1_2.parameters(), 'lr': LEARNING_RATE},
            {'params': model.classifier_layer_son2_1.parameters(), 'lr': LEARNING_RATE},
            {'params': model.classifier_layer_son2_2.parameters(), 'lr': LEARNING_RATE},
            {'params': Us, 'lr': LEARNING_RATE},
        ], lr=LEARNING_RATE / 10, momentum=momentum, weight_decay=l2_decay)

        try:
            source_data1, source_label1 = source1_iter.next()
        except Exception as err:
            source1_iter = iter(source1_loader)
            source_data1, source_label1 = source1_iter.next()
            
        try:
            source_data2, source_label2 = source2_iter.next()
        except Exception as err:
            source2_iter = iter(source2_loader)
            source_data2, source_label2 = source2_iter.next()
            
        try:
            target_data, __ = target_iter.next()
        except Exception as err:
            target_iter = iter(target_train_loader)
            target_data, __ = target_iter.next()
            
        if cuda:
            source_data1, source_label1 = source_data1.cuda(), source_label1.cuda()
            source_data2, source_label2 = source_data2.cuda(), source_label2.cuda()
            target_data = target_data.cuda()
        source_data1, source_label1 = Variable(source_data1), Variable(source_label1)
        source_data2, source_label2 = Variable(source_data2), Variable(source_label2)
        target_data = Variable(target_data)
        optimizer.zero_grad()

        cls_loss1, transfer_loss1,data_src1,outputs_data_src1 = model(source_data1,source_data2, target_data, source_label1,source_label2, mark=1)
        gamma = 2 / (1 + math.exp(-10 * (i) / (iteration) )) - 1
        loss1 = cls_loss1 + gamma * transfer_loss1
        
        cls_loss2, transfer_loss2,data_src2,outputs_data_src2 = model(source_data1,source_data2, target_data, source_label1,source_label2, mark=2)
        loss2 = cls_loss2 + gamma * transfer_loss2
        
        source_ids = range(2)
        hiddens = [data_src1,data_src2]
        hidden_labels = [source_label1,source_label2]
        ms_outputs = [ outputs_data_src1,outputs_data_src2]
        loss_entropy = []
        loss_moe = []

        for ind in source_ids:
            #support_ids = [ x for x in source_ids if x != ind ] 
            support_ids = [ x for x in source_ids] 
            source_alphas = [torch.exp(- mahalanobis_metric(hiddens[ind],hiddens[j].detach(),Us[j])) for j in source_ids ]
            source_alphas = softmax(source_alphas) # [ 32, 32]

            support_alphas = [ source_alphas[x] for x in support_ids ]
            support_alphas = softmax(support_alphas)
            # this is the mixture of expert classifier result with softmax result
            if cuda:
                source_alphas = [ alpha.cuda() for alpha in source_alphas ]
           
            output_moe_ind = sum([ alpha.unsqueeze(1) * F.softmax(ms_outputs[ind][id-1], dim=0) \
                                    for alpha, id in zip(support_alphas, support_ids) ]) 
            loss_moe.append(moe_criterion(torch.log(output_moe_ind), hidden_labels[ind])) #### meta_source and meta_target classification loss 

            source_alphas = torch.stack(source_alphas, dim=0)
            source_alphas = source_alphas.permute(1,0)
            loss_entropy.append(entropy_criterion(source_alphas))
        loss_moe = sum(loss_moe)
        loss_entropy = sum(loss_entropy)
        
        '''
        alpha1 = mahalanobis_metric(data_src2, data_src1.detach(), Us[0])
        alpha2 = mahalanobis_metric(data_src1, data_src2.detach(), Us[1])
        source_alphas = [alpha1, alpha2]
        source_alphas = softmax(source_alphas)
        loss_entropy = entropy_criterion(source_alphas[0]) + entropy_criterion(source_alphas[1])
        ''' 
        if cuda:
            source_alphas = [ alpha.cuda() for alpha in source_alphas ]
        '''
        output_moe_1 = source_alphas[1].unsqueeze(1) * F.softmax(outpus_data_src1, dim=1)
        output_moe_2 = source_alphas[0].unsqueeze(1) * F.softmax(outpus_data_src2, dim=1)

        loss_moe1 = moe_criterion(torch.log(output_moe_1), source_label1) #### meta_source and meta_target classification loss 
        loss_moe2 = moe_criterion(torch.log(output_moe_2), source_label2)
        loss_moe = loss_moe1 + loss_moe2
       ''' 
        loss = loss_moe + loss1 + loss2 + 0.01*loss_entropy
        
        loss.backward()
        optimizer.step()

        if i % log_interval == 0:
            print('Train source1 iter: {} [({:.0f}%)]\tsoft_Loss: {:.6f}\ttransfer_Loss: {:.6f}\tmoe_Loss: {:.6f}'.format(
                    i, 100. * i / iteration, cls_loss1.item(), transfer_loss1.item(),loss_moe.item()))

        if i % log_interval == 0:
            print('Train source2 iter: {} [({:.0f}%)]\tsoft_Loss: {:.6f}\ttransfer_Loss: {:.6f}\tent_Loss: {:.6f}'.format(
                    i, 100. * i / iteration, cls_loss2.item(), transfer_loss2.item(), loss_entropy.item()))
        
        if i % (log_interval * 4) == 0:
            t_correct = test(model,Us,source1_loader,source2_loader,target_test_loader)
            if t_correct > correct:
                correct = t_correct
            print(source1_name, source2_name, "to", target_name, "%s max correct:" % target_name, correct, "\n")
        
#        if i % 5000 == 0:
#            temp_model = nn.Sequential(model.sharedNet, model.sonnet1, model.sonnet2, model.avgpool)
#            torch.save(temp_model, osp.join("./result_time/model/",  source1+"_"+source2 +"_" + target+"_" +"iter_{:05d}_model.pth".format(i)))


def test(model,Us,source1_loader,source2_loader,target_test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    correct1 = 0
    correct2 = 0
    mu_S1,mu_S2 = domain_encoding(model,source1_loader,source2_loader)
    with torch.no_grad():    
        for data, target in target_test_loader:
            if cuda:
                data, target = data.cuda(), target.cuda()
            data, target = Variable(data), Variable(target)
            pred1, pred2,fea_son1,fea_son2 = model(data, mark = 0)
            
            mahalanobis_distances1 = (fea_son1 - mu_S1).mm(Us[0].mm(Us[0].t())).mm((fea_son1 - mu_S1).t())
            mahalanobis_distances2 = (fea_son2 - mu_S2).mm(Us[1].mm(Us[1].t())).mm((fea_son2 - mu_S1).t())
            alpha1 = mahalanobis_distances1.diag().sqrt().clamp(0.01, 2)
            alpha2 = mahalanobis_distances2.diag().sqrt().clamp(0.01, 2)
            
            #source_alphas = [alpha1, alpha2]
            source_alphas = [torch.exp(-alpha1), torch.exp(-alpha2)]
            source_alphas = softmax(source_alphas)
          
            pred1 = torch.nn.functional.softmax(pred1, dim=1)
            pred2 = torch.nn.functional.softmax(pred2, dim=1)
            
            #pred = source_alphas[0].unsqueeze(1)* pred1 + source_alphas[1].unsqueeze(1)*pred2
            #pred = source_alphas[1].unsqueeze(1)* pred1 + source_alphas[0].unsqueeze(1)*pred2
            pred = (pred1 + pred2) / 2
            test_loss += F.nll_loss(F.log_softmax(pred, dim=1), target).item()
    
            pred = pred.data.max(1)[1]
            correct += pred.eq(target.data.view_as(pred)).cpu().sum()
            pred = pred1.data.max(1)[1]
            correct1 += pred.eq(target.data.view_as(pred)).cpu().sum()
            pred = pred2.data.max(1)[1]
            correct2 += pred.eq(target.data.view_as(pred)).cpu().sum()
            #print("we are in test:",correct,correct1,correct2 )
            
        test_loss /= len(target_test_loader.dataset)
        print(target_name, '\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)\n'.format(
            test_loss, correct, len(target_test_loader.dataset),
            100. * float(correct) / len(target_test_loader.dataset)))
        print('\nsource1 accnum {}, source2 accnum {}'.format(correct1, correct2))
    return correct

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--dataset', default='Office-31', type=str,help='which dataset')
    parser.add_argument('--source1_name',type=str,default='amazon/images',help='address of image list for source1 dataset')
    parser.add_argument('--source2_name',type=str,default='dslr/images', help='address of image list for source2 dataset')
    parser.add_argument('--target_name', type=str,default='webcam/images', help='address of image list of target dataset')
    parser.add_argument('--lr', type=float, default=0.01, help="initial learning rate")
    parser.add_argument('--gpu_id', type=str, default='0', help='gpu_id')
    parser.add_argument('--log_interval', type=int, default=50, help='interval for output log')
    parser.add_argument('--rank_m', type=int, default=30, help='the rank of U')
    args = parser.parse_args()
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu_id
    
    source1_name = args.source1_name
    source2_name = args.source2_name
    target_name = args.target_name
    seed = 2
    cuda = True
    batch_size = 18
    log_interval = args.log_interval
    lr = args.lr
  
    torch.manual_seed(seed)
    if cuda:
        torch.cuda.manual_seed(seed)
    kwargs = {'num_workers': 4, 'pin_memory': True} if cuda else {}
    
    if args.dataset == 'Office-31':
        root_path = "../data/office31/"
        num_classes = 31 
        source1_loader = data_loader.load_clef_training(root_path, source1_name, batch_size, kwargs)
        source2_loader = data_loader.load_clef_training(root_path, source2_name, batch_size, kwargs)
        target_train_loader = data_loader.load_clef_training(root_path, target_name, batch_size, kwargs)
        target_test_loader = data_loader.load_clef_testing(root_path, target_name, batch_size, kwargs)  
        
    if args.dataset == "Image_clef":
        root_path = "../data/image-clef/"
        num_classes = 12
        source1_loader = data_loader.load_clef_training(root_path, source1_name, batch_size, kwargs)
        source2_loader = data_loader.load_clef_training(root_path, source2_name, batch_size, kwargs)
        target_train_loader = data_loader.load_clef_training(root_path, target_name, batch_size, kwargs)
        target_test_loader = data_loader.load_clef_testing(root_path, target_name, batch_size, kwargs)
    
    #model = models.MYDNN_COM_MOE(num_classes=31)
    model = models.MYDNN_MOE(num_classes=31)
 #   print(model)
    Us = []
    for i in range(2):
        U = torch.FloatTensor(256, args.rank_m*10)
        nn.init.xavier_uniform_(U)
        Us.append(U)
    Us = [ Variable(U.cuda(), requires_grad=True) for U in Us ]
    if cuda:
        model.cuda()
    torch.cuda.synchronize()
    start = time.time()
    train(source1_name,source2_name,target_name,model,Us,source1_loader,source2_loader,target_train_loader,target_test_loader,log_interval,lr)
    torch.cuda.synchronize()
    end = time.time()
    print("The end, test time：",end - start)
